# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Authors: Fernando Casanova <fcasanova@fluendo.com>,
#          Guillaume Emont <guillaume@fluendo.com>

import RDF

def add_shortcuts(model, source_node, target_node):
    """
    Find all statements relative to the source node and create equivalent
    statements relative to the target node.
    Statements can have source (and target) node as subject or as object.

    input:  A B C (A source, D target)

    output: D B C

    @param model: statements candidate for merging
    @type  model: C{RDF.Model}
    @param source_node: the source Node from which we will create the new
                   statements
    @type  source_node: C{RDF.Node}
    @param target_node: the Node for which we create new statements
    @type  target_node: C{RDF.Node}
    @return: the new model with new statements (shortcuts)
    @rtype:  C{RDF.Model}
    """
    for statement in model.find_statements(RDF.Statement(source_node, None,
                                                         None)):
        new_statement = RDF.Statement(target_node,
                                      statement.predicate,
                                      statement.object)
        model.add_statement(new_statement)

    for statement in model.find_statements(RDF.Statement(None, None,
                                                         source_node)):
        new_statement = RDF.Statement(statement.subject,
                                      statement.predicate,
                                      target_node)
        model.add_statement(new_statement)

    return model


class MetadataSource(object):
    """
    Standard API for metadata source, to wrap around any provider of metadata.

    @cvar  handled_classes: uris to classes about which this source can give
                            information
    @ctype handled_classes: C{list} of C{str}
    """
    name = None
    handled_classes = []
    confidence = 1.0

    def __init__(self, manager=None, **kw):
        super(MetadataSource, self).__init__(**kw)
        self._manager = manager

    def find_statements(self, subject, predicate, object):
        """
        Fetch metadata corresponding to the subject, predicate and object in
        parameter. Either subject or object should be None.

        It is the responsibility of the source to return consistent results,
        in particular, if it can find several result candidates that
        contradict each other (e.g. as a result to a search query to a remote
        database), the source should select one of them.

        @param subject: the subject for which we look for information
        @type subject: C{RDF.Node} or C{NoneType}
        @param predicates: the predicate in which we are interested
        @type predicates: C{list} of C{RDF.Node}
        @param object: the object for which we look for information
        @type object: C{RDF.Node} or C{NoneType}

        @return: a deferred that will be fired with all the statements found
                 while performing the necessary request, these might include
                 unrelated statements, so that the caller can cache them.
                 Alternatively, the deferred can be fired with C{None},
                 meaning that no result could be found.
        @rtype:  C{Deferred} to be fired with either an iterable of
                 C{RDF.Statement} or C{NoneType}
        """
        raise NotImplementedError()

