# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Fernando Casanova <fcasanova@fluendo.com>


import RDF
rdfs = RDF.NS('http://www.w3.org/2000/01/rdf-schema#')

from core.media_scanner.rdf_model import RDFModel
from plugins.name_parser.metadata_source import NameParserDataSource

from twisted.internet import defer

class Identificator(object):
    debug = True
    _mime_type = None
    _data_sources = []
    _pre_treatment_data_sources = []
    all_or_nothing = True

    def register_data_source(self, data_source):
        self._data_sources.append(data_source)

    def _pre_treat_model(self, model):
        # Overridable
        return self.identify_model(model, self._pre_treatment_data_sources)

    def _get_sorted_data_sources(self, model, data_sources):
        # Overridable
        return sorted(data_sources, \
                      key=lambda data_source: \
                                 data_source.confidence, \
                      reverse=True)

    def _post_pre_treat(self, result, model, data_sources):
        def filter_results(result):
            search_results = []
            for (success, value) in result:
                if success and len(value) != 0:
                    for i in value:
                        search_results.append(i)
            return defer.succeed(search_results)

        ordered_data_sources = self._get_sorted_data_sources(model, data_sources)

        search_results = []
        list_defers = []
        for data_source in ordered_data_sources:
            dfr = data_source.get_best_match(model)
            list_defers.append(dfr)

        # FIXME WHY DO WE NEED ANOTHER LIST OF DEFERS IF WE WANT TO PARALLELIZE!!!

        dl = defer.DeferredList(list_defers)
        dl.addCallback(filter_results)
        dl.addCallback(self._merge_search_results_into_model, model)
        return dl

    def identify_model(self, model, data_sources=None):
        if data_sources is None:
            data_sources = self._data_sources
            dfr = self._pre_treat_model(model)
        else:
            dfr = defer.succeed(True)

        dfr.addCallback(self._post_pre_treat, model, data_sources)
        return dfr

    def fill_model(self, model):
        # Retrieve Data Source who identified
        query_statement = RDF.Statement(None, rdfs['wasIdentifiedBy'], None)
        stream = model.find_statements(query_statement)
        data_source_to_query = None
        for statement in stream:
            data_source_to_query = statement.object
        return data_source_to_query is not None

        output = (not data_source_to_query is None)

        for data_source in self._data_sources:
            if data_source.name == str(data_source_to_query):
                output = output or data_source.fill_model(model)
        return output

    def _merge_search_results_into_model(self, search_results, model):
        # generic implementation
        ordered_results = sorted(search_results, key=self.compute_search_result_score, reverse=True)

        something_appended = False
        first_result = True

        for result in ordered_results:
            # Start by going through the contexts, from highest score
            # to lowest score.

            statements = result.get_all_statements(model.media_id, [])
            for statement in statements:
                if model.has(statement):
                    # already have
                    pass
                else:
                    # Get a statement that the new context has NOT in common with
                    # result_model then append it
                    model.append(statement)
                    something_appended = True

            if self.debug:
                data_source_result = ""
                if result.data_source.name != 'nameparser':

                    query_statement = RDF.Statement(None, rdfs['file_name'], None)
                    stream = model.find_statements(query_statement)
                    for statement in stream:
                        file_name = unicode(statement.object)

                    query_statement = RDF.Statement(None, rdfs['prelabel'], None)
                    stream = model.find_statements(query_statement)
                    for statement in stream:
                        query_string = unicode(statement.object)

                    query_statement = RDF.Statement(None, rdfs['label'], None)
                    stream = result.model.find_statements(query_statement)
                    for statement in stream:
                        label = unicode(statement.object)

                    if first_result:
                        data_source_result = "FILE: \033[0;34m%s\033[m" % file_name
                    else:
                        data_source_result = "\033[0;31m%s\033[m" % ("")
                    while len(data_source_result) < 90:
                        data_source_result += " "
                    data_source_result += "%10s score %.2f" % (result.data_source.name, self.compute_search_result_score(result))
                    data_source_result += " [%s] " % (query_string)
                    data_source_result += "[\033[0;33m%s\033[m]" % (label)
                    if first_result and self.compute_search_result_score(result) > 0.75:
                        data_source_result += " [\033[0;32mOK\033[m]"
                        first_result = False
                    else:
                        data_source_result += " [\033[0;31mKO\033[m]"
            if self.debug and data_source_result != '':
                print data_source_result

            if something_appended:
                return defer.succeed(model)

        return defer.succeed(model)


    def compute_search_result_score(self, search_result):
        # overriddable
        # uses:
        # - search_result.model.quantity_of_information
        # - search_result.data_source.confidence
        # - search_result.confidence
        score = 1.0 * search_result.data_source.confidence + \
                2.0 * search_result.confidence# + \
#                0.5 * search_result.model.size()
        return score / 3.0

def parse_file(file):
    storage = RDF.MemoryStorage()
    model = RDFModel(storage)
    if model is None:
        raise "[PF] New RDF.model failed"

    print "[PF] Parsing URI (file)", file
    uri = RDF.Uri(string = "file:" + file)

    parser = RDF.Parser('raptor')
    if parser is None:
        raise "[PF] Failed to create RDF.Parser raptor"

    count = 0
    for s in parser.parse_as_stream(uri, uri):
        model.add_statement(statement=s)
        count += 1
    print "[PF] Parsing added", count, "statements"
    return model

class VideoIdentificator(Identificator):
    _mime_type = 'video'
    _pre_treatment_data_sources = [NameParserDataSource()]

    def _get_sorted_data_sources(self, model, data_sources):
        is_tv_show = False
        query_statement = RDF.Statement(None, rdfs['might_be'], RDF.Node('tvshow'))
        stream = model.find_statements(query_statement)
        for statement in stream:
            is_tv_show = True

        my_data_sources = []

        index_of_tvdb = 0
        for i in range(0, len(data_sources)):
            if data_sources[i].name == 'thetvdb':
                index_of_tvdb = i
            my_data_sources.append(data_sources[i])

        if is_tv_show:
            thetvdb = my_data_sources.pop(index_of_tvdb)
            my_data_sources.insert(0, thetvdb)

        return sorted(my_data_sources, \
                      key=lambda data_source: \
                                 data_source.confidence, \
                      reverse=True)

if __name__ == '__main__':
    myModel = parse_file('input.film.rdf')
    myId = VideoIdentificator()
    myDataSource = DBPediaDataSource()
    myId.register_data_source(myDataSource)
    myId.identify_model(myModel)
