# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Guillaume Emont <guillaume@fluendo.com>

import RDF
import Levenshtein

from twisted.internet import defer

from core.iteration import sequential_iteration, parallel_iteration, \
         stop_iteration

rdf = RDF.NS('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
rdfs = RDF.NS('http://www.w3.org/2000/01/rdf-schema#')

def stream_wrapper(stream):
    # FIXME: should be put somewhere in utils
    for element in stream:
        yield element

class Scorer(object):
    weight = 1.0
    def __init__(self, manager=None, **kw):
        super(Scorer, self).__init__(**kw)
        self.manager = manager

    def get_score(self, statement, source):
        """
        Return a score between 0 and 1 or None, through a deferred
        """
        return defer.succeed(None)


class LevenshteinScorer(Scorer):
    weight = 2.0
    described = rdfs['describedBy']
    label = rdfs['label']
    clean_name = rdfs['clean-name']

    def get_score(self, statements, source):
        # we score by levenshtein-comparing the label held in statements and
        # a clean name we ask to the manager
        #FIXME: async!
        label_statements = []
        described_statement = None
        for statement in statements:
            if statement.predicate == self.label:
                label_statements.append(statement)
            elif statement.predicate == self.described:
                described_statement = statement

        def compute_ratio(node, label_node):
            dfr = self.manager.find_statements(node, self.clean_name, None)
            def return_ratio(statements):
                clean_name_statement = statements.next()
                clean_name = unicode(clean_name_statement.object)
                return Levenshtein.ratio(clean_name, unicode(label_node))

            dfr.addCallback(return_ratio)
            return dfr

        for statement in label_statements:
            if statement.subject == described_statement.subject:
                return compute_ratio(statement.subject, statement.object)

        return defer.succeed(None)


class SourceScorer(Scorer):
    def get_score(self, statement, source):
        return defer.succeed(source.confidence)


class MetadataManager(object):
    # FIXME: modify the API to avoid reflecting the RDF implementation choice
    # (redland)
    """
    DOCME
    @ivar model: The model where all the cached data is stored
    @type model: C{RDF.Model}
    """
    class_predicate = rdf['type']

    def __init__(self, storage_file=None, **kw):
        super(MetadataManager, self).__init__(**kw)
        if storage_file is None:
            storage_file = 'metadata'
        storage = RDF.Storage(storage_name='hashes', name=storage_file,
                              options_string="hash-type='bdb',index-predicates='yes'")

        self.model = RDF.Model(storage)

        # class uri->list of (source, merger) tuples. We only discriminate on
        # that for now to chose the sources we ask for data.
        self._class_sources = {}
        # source->merger
        self._mergers = {}

        # list of Scorer objects
        self._scorers = []

    def execute(self, query):
        return self.model.execute(query)

    def find_statements(self, subject, predicate, object,
                         first_match=False, use_cache=True):
        """
        DOCME
        @param first_match: whether we stop trying to find a result at the
                            first match, if the data is not in the cache
        @type first_match: C{bool}
        @param use_cache: request data from the providers only if the cache
                          doesn't have it. Note that if you specify
                          use_cache=False, some data from the cache might
                          still be used by the data source.
        @type use_cache: C{bool}
        """
        # return an iterable (iterator, actually) of statements corresponding
        # to the request
        query_statement = RDF.Statement(subject, predicate, object)

        if use_cache:
            stream = self.model.find_statements(query_statement)
        if not use_cache or stream.end(): # stream.end() means empty stream
            return self._lookup_statements(subject, predicate, object,
                                           first_match)

        return defer.succeed(stream_wrapper(stream))

    def remove_statement(self, statement):
        self.model.remove_statement(statement)

    def add_statements(self, *statements):
        for statement in statements:
            self.model.append(statement)

    def add_source(self, source, merger):
        def add_value_for_keys(d, keys, value):
            for key in keys:
                if key not in d:
                    d[key] = []
                d[key].append(value)

        add_value_for_keys(self._class_sources, source.handled_classes, source)
        self._mergers[source] = merger

    def add_scorer(self, scorer):
        self._scorers.append(scorer)

    def _lookup_statements(self, subject, predicate, object, first_match):
        scores = []

        if first_match:
            # we want to do things in order, sequentially, and stop at the
            # first match
            sources = self._get_ordered_sources(subject, predicate, object)
            iteration = sequential_iteration
        else:
            # we want to get all possible results as fast as possible
            sources = self._get_relevant_sources(subject, predicate, object)
            iteration = parallel_iteration

        def score_statements(statements, source):
            if statements:
                dfr = self._score(statements, source)
                def save_score(score):
                    scores.append((score, statements, source))
                dfr.addCallback(save_score)

                if first_match:
                    dfr.addCallback(stop_iteration)

                return dfr

        def find_and_score_statements(source):
            dfr = source.find_statements(subject, predicate, object)
            dfr.addCallback(score_statements, source)
            return dfr

        def find_best(result):
            sorted_scores = sorted(scores, key=lambda x: x[0])
            return sorted_scores[0]

        def cache_and_return(result):
            score, statements, source = result
            dfr = self._merge_statements(source, statements)
            dfr.addCallback(lambda x: statements)
            return dfr

        dfr = iteration(sources, find_and_score_statements)
        dfr.addCallback(find_best)
        dfr.addCallback(cache_and_return)

        return dfr

    def _merge_statements(self, source, statements):
        return self._mergers[source].merge_statements(statements, source)

    def _score(self, statements, source):
        def get_score(scorer):
            dfr = scorer.get_score(statements, source)
            def return_weight_and_score(score):
                return (scorer.weight, score)
            dfr.addCallback(return_weight_and_score)
            return dfr


        def compute_result(results):
            weight_sum = 0.0
            score_sum = 0.0
            for success, result in results:
                if not success:
                    continue
                weight, score = result
                weight_sum += weight
                score_sum += weight * score

            return score_sum / weight_sum

        dfr = parallel_iteration(self._scorers, get_score)
        dfr.addCallback(compute_result)

        return dfr

    def _get_relevant_sources(self, subject, predicate, object):
        if subject is None:
            known_object = object
        else:
            known_object = subject
        klass = self.model.get_target(known_object, self.class_predicate)
        return self._class_sources.get(klass, [])

    def _get_ordered_sources(self, subject, predicate, object):
        sources = self._get_relevant_sources(subject, predicate, object)
        def source_key(source):
            return source.confidence
        return sorted(sources, key=source_key, reverse=True)
