# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Authors: Fernando Casanova <fcasanova@fluendo.com>
#          Guillaume Emont <guillaume@fluendo.com>

from core.media_scanner.data_source import MetadataSource
from core.queue import CancellableQueue

import libxml2, libxslt
from xml.dom import minidom

import RDF
import pkg_resources
import re

from twisted.internet import defer, task
from twisted.web.client import getPage


class InfoGetter(object):
    def do_request(self, uri):
        """
        Return an a deferred that will be fired with an iterable of
        C{RDF.Model} with the information we can get from uri.
        """
        # FIXME: I don't like the name of this method
        raise NotImplementedError()


class XmlInfoGetter(InfoGetter):
    """
    @ivar xslt_dict: mapping of uri regexps to tuples containing:
     - an xslt file reference in the form "module:resource"
     - parameters for the xslt
    @itype xslt_dict: C{dict} of C{str} -> (C{str}, C{dict})
    """
    # FIXME: should we split http-specific stuff and xml-specific stuff?

    TIMEOUT = 5.0
    max_number_of_connexions = 20

    def __init__(self, xslt_dict=None):
        self.xslt_dict = xslt_dict
        self.connexion_queue = CancellableQueue(self.max_number_of_connexions)

    def do_request(self, uri):
        """
        Uri should be matched by one of the regexps in self.xslt_dict.keys()
        """
        def get_xslt_string(result):
            # result is what we got from _get_xslt_entry()
            ref, params = result
            module, path = ref.split(':')
            xslt_string = pkg_resources.resource_string(module, path)
            return xslt_string, params

        def get_xml_string(result):
            xslt_string, params = result
            dfr = self._get_page(uri)
            dfr.addCallback(lambda xml_string: \
                            (xml_string, xslt_string, params))
            return dfr

        def apply_stylesheet(result):
            return self._apply_stylesheet(*result)

        dfr = self._get_xslt_entry(uri)
        dfr.addCallback(get_xslt_string)
        dfr.addCallback(get_xml_string)
        dfr.addCallback(apply_stylesheet)
        dfr.addCallback(self._make_models)

        return dfr

    def _get_xslt_entry(self, uri):
        for (pattern, (xslt_ref, xslt_params)) in self.xslt_dict.iteritems():
            if re.match(pattern, uri):
                return defer.succeed((xslt_ref, xslt_params))

        raise ValueError()

    def _apply_stylesheet(self, xml_doc, xslt_string, xslt_args):
        styleDoc = libxml2.parseDoc(xslt_string)
        style = libxslt.parseStylesheetDoc(styleDoc)

        doc = libxml2.parseDoc(xml_doc)

        result = style.applyStylesheet(doc, xslt_args)

        if len(result.content) == 0:
            return None

        res = style.saveResultToString(result)

        style.freeStylesheet()
        doc.freeDoc()
        result.freeDoc()

        return defer.succeed(res)

    def _make_models(self, rdf_xml_string):
        parser = RDF.Parser(name="rdfxml")

        result_parsed = minidom.parseString(rdf_xml_string)
        results = result_parsed.getElementsByTagName("rdf:RDF")

        output = []

        def make_model(rdf_string):
            model = RDF.Model()
            # FIXME: shouldn't we put a real uri here?
            parser.parse_string_into_model(model, rdf_string, RDF.Uri('fake'))
            return defer.succeed(model)

        def add_to_output(model):
            output.append(model)

        def iterate_results():
            for result_DOM in results:
                model_dfr = make_model(result_DOM.toxml())
                model_dfr.addCallback(add_to_output)
                yield model_dfr

        dfr = task.coiterate(iterate_results())
        dfr.addCallback(lambda x: output)

        return dfr

    def _get_page(self, url):
        def treat_error(failure):
            print "ERROR _get_page (probable timeout for %s)" % uri
            return None

        dfr = self.connexion_queue.enqueue(getPage, url, timeout=self.TIMEOUT)
        dfr.addErrback(treat_error)
        return dfr


class GetterDataSource(MetadataSource):
    def __init__(self, getter=None, **kw):
        super(GetterDataSource, self).__init__(**kw)
        self._getter = getter

    def find_statements(self, subject, predicate, object):
        # FIXME: error handling
        dfr = self._query_to_uri(subject, predicate, object)
        dfr.addCallback(self._getter.do_request)
        dfr.addCallback(self._select_result, subject, predicate, object)

        return dfr

    def _query_to_uri(self, subject, predicate, object):
        """
        This method should be implemented in subclasses

        @return: the uri to use to get information to answer the query made by
        (subject, predicate, object)
        @rtype:  C{Deferred} that will be fired with a C{str}
        """
        raise NotImplementedError()

    def _select_result(self, models, subject, predicate, object):
        """
        DOCME
        """
        raise NotImplementedError()

