# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Guillaume Emont <guillaume@fluendo.com>

import random, os

import RDF
from twisted.trial.unittest import TestCase
from twisted.internet import defer

from core.media_scanner.metadata_manager import MetadataManager, \
         SourceScorer, Merger

from core.media_scanner.data_source import MetadataSource

test_namespace = RDF.NS('http://foo.bar/bar#')
test_class = test_namespace['klass1']

class TestSource(MetadataSource):
    name = "test source"
    handled_classes = [test_class]
    confidence = 1.0
    def __init__(self, **kw):
        super(TestSource, self).__init__(**kw)
        self.result = []

    def find_statements(self, subject, predicate, object):
        return defer.succeed(self.result)

class SimpleMerger(Merger):
    """Merge everything, without any check."""
    def merge_statements(self, statements, source):
        self.manager.add_statements(*statements)
        return defer.succeed(None)

class MetadataManagerTest(TestCase):
    size = 50
    def setUp(self):
        self.manager = MetadataManager('test_metadata')

    def tearDown(self):
        del self.manager.model
        del self.manager
        for suffix in ('sp2o', 'po2s', 'so2p', 'p2so'):
            os.remove("test_metadata-%s.db" % suffix)


class MetadataManagerSimpleTest(MetadataManagerTest):
    """
    Simple tests with a MetadataManager with no source. This tests mostly the
    "store" aspect of the MetadataManager.
    """
    def setUp(self):
        super(MetadataManagerSimpleTest, self).setUp()
        self.nodes = [RDF.Node(blank=str(i)) for i in xrange(self.size)]
        self.predicates = [test_namespace[str(i)]
                                for i in xrange(self.size)]
        self.statements = self._make_random_statements(self.nodes,
                                                        self.predicates)

    def _make_random_statements(self, nodes, predicates):
        result = []
        for i in xrange(2*self.size):
            subject = self.nodes[random.randint(0, self.size-1)]
            object = self.nodes[random.randint(0, self.size-1)]
            predicate = self.predicates[random.randint(0, self.size-1)]
            statement = RDF.Statement(subject, predicate, object)
            if statement not in result:
                result.append(statement)
        return result

    @defer.inlineCallbacks
    def test_add(self):
        statement = self.statements[0]
        self.manager.add_statements(statement)
        statements = yield self.manager.find_statements(None, None, None)
        statements = list(statements)
        self.failUnlessEqual(len(statements), 1)
        self.failUnlessEqual(statements[0], statement)

    def _compare_sets(self, set1, set2):
        # we need this because we want to consider equal two sets that have
        # equal elements, not only two sets that have the same elements (the
        # same instances)
        if len(set1) != len(set2):
            return False

        for elem in set1:
            if elem not in set2:
                return False

        return True

    @defer.inlineCallbacks
    def test_add_multi(self):
        self.manager.add_statements(*self.statements)
        statements = yield self.manager.find_statements(None, None, None)
        statements = list(statements)
        self.failUnless(self._compare_sets(statements, self.statements))

    @defer.inlineCallbacks
    def test_remove(self):
        if len(self.statements) < 2:
            return
        self.manager.add_statements(*self.statements)
        self.manager.remove_statement(self.statements[0])
        statements = yield self.manager.find_statements(None, None, None)
        statements = list(statements)
        self.failUnless(self._compare_sets(statements, self.statements[1:]))

class MetadataManagerSimpleSourceTest(MetadataManagerTest):
    def setUp(self):
        super(MetadataManagerSimpleSourceTest, self).setUp()
        self.source = TestSource()
        self.manager.add_source(self.source, SimpleMerger(manager=self.manager))
        self.manager.add_scorer(SourceScorer())
        self.cool_node = RDF.Node(blank='cool')
        class_statement = RDF.Statement(self.cool_node,
                                        self.manager.class_predicate,
                                        test_class)
        self.manager.add_statements(class_statement)

    @defer.inlineCallbacks
    def test_simple_lookup(self):
        result_statement = RDF.Statement(self.cool_node,
                                         test_namespace['foo'],
                                         RDF.Node(blank='bar'))
        self.source.result = [result_statement]
        statements = yield self.manager.find_statements(self.cool_node,
                                                        test_namespace['foo'],
                                                        None)
        self.failIfEqual(statements, None)
        statements = list(statements)
        self.failUnlessEqual(len(statements), 1)
        self.failUnlessEqual(statements[0], result_statement)

    @defer.inlineCallbacks
    def test_lookup_and_caching(self):
        yield self.test_simple_lookup()
        old_result_statement = RDF.Statement(self.cool_node,
                                         test_namespace['foo'],
                                         RDF.Node(blank='bar'))
        result_statement = RDF.Statement(self.cool_node,
                                         test_namespace['foo'],
                                         RDF.Node(blank='baz'))
        self.source.result = [result_statement]

        # this should give the cached result
        statements = yield self.manager.find_statements(self.cool_node,
                                                        test_namespace['foo'],
                                                        None)
        self.failIfEqual(statements, None)
        statements = list(statements)
        self.failUnlessEqual(len(statements), 1)
        self.failUnlessEqual(statements[0], old_result_statement)

        # now, with caching disabled, we should get the new result returned by
        # the source:
        statements = yield self.manager.find_statements(self.cool_node,
                                                        test_namespace['foo'],
                                                        None,
                                                        use_cache=False)
        self.failIfEqual(statements, None)
        statements = list(statements)
        self.failUnlessEqual(len(statements), 1)
        self.failUnlessEqual(statements[0], result_statement)
