
class MetadataSource(object):
    """
    Standard API for metadata source, to wrap around any provider of metadata.

    @cvar object_fetch_predicates: Predicates which can be used with
        fetch_metadata() with a known subject (and None as object)
    @ctype object_fetch_predicates: C{list} of C{RDF.Node}
    @cvar subject_fetch_predicates: Predicates which can be used with
        fetch_metadata() with a known object (and None as subject)
    @ctype subject_fetch_predicates: C{list} of C{RDF.Node}
    @cvar searchable_classes: uris to classes that search() can handle
    @ctype searchable_classes: C{list} of C{str}
    """
    object_fetch_predicates = []
    subject_fetch_predicates = []
    searchable_classes = []

    def search(self, search_term, result_class):
        """
        Perform a search

        @param search_term: our search query (typically, a title)
        @type search_term: urlencoded C{str}
        @param result_class: a uri to a class in an ontology, the results of
                the search will be of that class or a subclass of it
        @type result_class: C{str}

        @return: list of results, each containing info about one potential
        result
        @rtype C{list} of C{SearchResult}
        """
        raise NotImplementedError

     def fetch_metadata(self, model, subject, predicates, object):
        """
        Fetch metadata corresponding to the subject, predicates and object in
        parameter. Either subject or object should be None.

        @param model: the model in which to put the statements we find.
        @type model: C{RDF.Model}
        @param subject: the subject for which we look for information
        @type usbject: C{RDF.Node} or C{NoneType}
        @param predicates: the predicates in which we are interested
        @type predicates: C{list} of C{RDF.Node}
        @param object:
        @type object:
        """
        raise NotImplementedError

class MetadataManager(object):
    def __init__(self):
        # super()
        self._sources = []

    def search(self, search_term, result_class):
        pass

    def fetch_metadata(self, model, subject, predicates, object):
        # this should handle the conversion of the subject/object from
        # something that make sense locally to something that makes sense for
        # the data source.
        pass

    def add_source(self, source):
        self._sources.append(source)

# Example pseudo-implementation:

class XmlInfoGetter(object):
    def __init__(self, xslt_dict=None):
        """
        @param xslt_dict: mapping of uri regexps to xslt strings
        @type xslt_dict: C{dict} of C{str} -> C{str}
        """
        self.xslt_dict = xslt_dict

    def do_request(self, model, uri):
        """
        Fill model with the information we can get from uri (an xml document),
        after having applied the right stylesheet to make an XMLRDF document of
        it.

        uri should be matched by one of the regexps in self.xslt_dict.keys()
        """
        pass

class XsltSource(MetadataSource):
    """
    @cvar searches: mapping of classes for which we support search and the
    corresponding uri
    @ctype searches: C{dict} of C{str} -> C{tuple} of (C{str}, C{str})
    """

    searches = {}
    object_fetches = {}
    subject_fetches = {}

    def __init__(self, *args, **kw):
        super(XsltSource, self).__init__(*args, **kw)
        self._getter = XmlInfoGetter()

    def _make_search_results(self, model, result_class):
        # Use the info from model to generate separate SearchResult instances
        # by filtering the instances (hint: that consists in finding the
        # disjoint subgraphs of the graph).

    def search(self, search_term, result_class):
        model = RDF.Model()
        uri = self.searches[result_class]
        self._getter.do_request(model, uri)
        return self._make_search_results(model)

    def fetch_metadata(self, model, subject, predicates, object):
        if (subject, object) == (None, None):
            raise ValueError()
        if object is None:
            for predicate in predicates:
                uri = self.object_fetches[predicate]
                self._getter.do_request(model, uri)
        elif subject is None:
            for predicate in predicates:
                uri = self.subject_fetches[predicate]
                self._getter.do_request(model, uri)
        else:
            raise ValueError()


class TheMovieDBSource(XsltSource):
    API_KEY = "kjezjakelkjzajlkzaje"

    movie_search_uri = 'http://..../Movie.search/.../%s/' % API_KEY + '%s'
    movie_info_uri = 'http://.../Movie.getInfo/.../%s/' % API_KEY + '%s'
    person_search_uri = 'http://..../Person.search/.../%s/' % API_KEY + '%s'
    person_info_uri = 'http://.../Person.getInfo/.../%s/' % API_KEY + '%s'

    movie_xslt = "truc/movie.xsl"
    person_xslt = "truc/person.xsl"

    object_fetches = {
        rdfs['label']: movie_info_uri,
        rdfs['score']: movie_info_uri,
        rdfs['alternative_name']: movie_info_uri,
        cool_person_ontology_ns['birthday']: person_info_uri,
        blabla_ns['alias']: person_info_uri,
    }
    object_fetch_predicates = object_fetches.keys()

    subject_fetches = {
        # FIXME: check if we are guaranteed of unique predicates like here; if
        # it can be something like 'movie' for everybody, we need to restrict
        # this by adding the class of the subject here.
        blah['performance-movie']: movie_info_uri,
        blah['director-movie']: movie_info_uri,
        blah['performance-actor']: person_info_uri,
        blah['direction-director']: person_info_uri,
    }
    subject_fetch_predicates = subject_fetches.keys()

    searches = {
        cool_ontology_ns['Movie']: movie_search_uri,
        cool_ontology_ns['Person']: person_search_uri,
    }
    searchable_classes = searches.keys()

    def __init__(self, *args, **kw):
        # super( ...).__init__(...)
        movie_xslt_string = open(movie_xslt).read().replace('API_KEY', self.API_KEY)
        person_xslt_string = open(person_xslt).read().replace('API_KEY', self.API_KEY)
        self._getter.xslt_dict['.*Movie.*'] = movie_xslt_string
        self._getter.xslt_dict['.*Person.*'] = person_xslt_string


# note on the performance class as I imagine it here (it probably exist in some
# ontology, or something similar):
# Performance:
#   movie
#   actor
#   role
#
# A actor Silvester Stallone
# A role Rocky Balboa
# A movie Rocky IV
#
# Idem we can have:
# Direction:
#   movie
#   director
# Production:
#   movie
#   producer


# Examples of use

###########
# media scanner
###########

# these are to be computed as the union of the predicates declared by the users
# of the result (typically: UI controllers). They should declare the data they
# need right away and cannot afford to fetch dynamically.
# E.g. a UI that want to allow the browsing of music by artists need the media
# scanner to find these at scanning time, and therefore would do something like:
#   audio_track_needed_metadata_as_subject.append(blah['author'])
# It then can safely do a SPARQL query to get all the authors, and then for
# each author to get all the tracks.

tv_show_needed_metadata_as_subject = []
tv_show_needed_metadata_as_object = []
movie_needed_metadata_as_subject = []
movie_needed_metadata_as_object = []
audio_track_needed_metadata_as_subject = []
audio_track_needed_metadata_as_object = []

def simple_scan_media(base_uri, manager):
    crawler = Crawler()
    model = RDF.Model()
    for uri in crawler.crawl():
        file_node = RDF.Node(RDF.Uri(uri))
        manager.fetch_metadata(model,
                               file_node,
                               [blah['type']], # something that would be one of (video,audio,image)
                               None)
        found_type = model.get_target(file_node, blah['type'])
        if found_type.literal_value == 'video':
            results = manager.search(make_search_term(uri), blah['video'])
            # get best result and put it in model
            # ....
            if is_of_type(model, file_node, blah['tv_show']):
                manager.fetch_metadata(model, file_node, tv_show_needed_metadata_as_subject, None)
                manager.fetch_metadata(model, None, tv_show_needed_metadata_as_object, file_node)
        elif found_type.literal_value == 'audio':
            pass

############
# UI
############

# TODO
# The basic idea is that the UI can do SPARQL query for stuff that is
# guaranteed to be fetched at scan time (see the lists above
# simple_scan_media()), and can use manager.fetch_metadata() for everything
# else; I think its needs should be covered that way.



########################
# OLD ARCHI

class RDFModel
    quantity_of_information
    mime_type
    type
    model

class SearchResult
    data_source
    confidence
    rdf_model

class DataSource
    name
    mime_type
    confidence
    media_id

    def get_best_match(model):
        # generic implementation
        search_parameter = self.extract_search_parameter(model)

        search_results = self._search(search_parameter)
        search_results.sort(key="confidence")
        search_result = search_results[0]
        return search_result

    def _search(search_parameter):
        # generic implementation
        response = self.execute_search(search_parameter)

        for item in response:
            result = SearchResult()
            result.data_source = self
            result.model = item
            result.confidence = self.compute_confidence(item, search_parameter)
        return search_results

    def compute_confidence(model, search_parameter):
        # overriddable
        output = self.extract_search_parameter(model)
        return Levenshtein.ratio(search_parameter, output)

    def execute_search(search_parameter):
        # to be overridden

    def extract_search_parameter(model)
        # to be overridden
        search_parameter = model.label
        self.media_id = model.id
        return search_parameter

class Identificator
    mime_type
    data_sources

    def identify_model(model):
        # generic implementation
        search_results = []
        for data_source in data_sources:
            search_result = data_source.get_best_match(model)
            search_results.append(search_result)

        self._merge_search_results_into_model(search_results, model)

        return model.type

    def _merge_search_results_into_model(search_results, model):
        # generic implementation
        [...]
        ordered_results = sorted(search_results, key=self.compute_search_result_score)
        [...]

    def compute_search_result_score(search_result):
        # overriddable
        # uses:
        # - search_result.model.quantity_of_information
        # - search_result.data_source.confidence
        # - search_result.confidence
        score = search_result.data_source.confidence + 2.0 * search_result.confidence
        return score
